# Some guidelines:
#
# Put code in logical blocks using methods. That way they can be reused
# and also show up in the "source structure" of most IDE's.
#
# Colors should be converted to the physical colorspace in the fragment
# shader (because you generally want to interolate in srgb). Other than
# that, the color should be turned to physical using srgb2physical()
# as soon as reasonably possible. For clarity it helps to add a "_srgb"
# suffix to colors not yet converted. Most colors on a uniform can also
# be considered srgb.
#
# The function names are left camelCase, many come from ThreeJS and this makes it
# easy to look them up. In general, the variables are snake_case.
#
# The PI and RECIPROCAL scale factors for the lights is a complex story
# related to artist-friendly colors, historic design choices, and
# undoing the PI factor in other places. We just try follow whay ThreeJS
# is doing here.


class Shaderlib:
    def light_deps_basic(self):
        return """
        // TODO smoothstep and saturate probably exists when Naga gets updated
        fn _smoothstep( low : f32, high : f32, x : f32 ) -> f32 {
            let t = clamp( ( x - low ) / ( high - low ), 0.0, 1.0 );
            return t * t * ( 3.0 - 2.0 * t );
        }
        fn _saturate( x: f32) -> f32 {
            return clamp(x, 0.0, 1.0);
        }
        fn getDistanceAttenuation(light_distance: f32, cutoff_distance: f32, decay_exponent: f32) -> f32 {
            var distance_falloff: f32 = 1.0 / max( pow( light_distance, decay_exponent ), 0.01 );
            if ( cutoff_distance > 0.0 ) {
                distance_falloff *= pow2( _saturate( 1.0 - pow4( light_distance / cutoff_distance ) ) );
            }
            return distance_falloff;
        }
        fn getSpotAttenuation( cone_cosine: f32, penumbra_cosine: f32, angle_cosine: f32 ) -> f32 {
            return _smoothstep( cone_cosine, penumbra_cosine, angle_cosine );
        }
        fn getAmbientLightIrradiance( ambientlight_color: vec3<f32> ) -> vec3<f32> {
            let irradiance = ambientlight_color;
            return irradiance;
        }
        struct IncidentLight {
            color: vec3<f32>,
            direction: vec3<f32>,
            visible: bool,
        };
        struct ReflectedLight {
            direct_diffuse: vec3<f32>,
            direct_specular: vec3<f32>,
            indirect_diffuse: vec3<f32>,
            indirect_specular: vec3<f32>,
        };
        struct GeometricContext {
            position: vec3<f32>,
            normal: vec3<f32>,
            view_dir: vec3<f32>,
        };
        $$ if num_dir_lights > 0
        fn getDirectionalLightInfo( directional_light: DirectionalLight, geometry: GeometricContext ) -> IncidentLight {
            var light: IncidentLight;
            light.color = srgb2physical(directional_light.color.rgb) * directional_light.intensity;
            light.direction = -directional_light.direction.xyz;
            light.visible = true;
            return light;
        }
        $$ endif
        $$ if num_point_lights > 0
        fn getPointLightInfo( point_light: PointLight, geometry: GeometricContext ) -> IncidentLight {
            var light: IncidentLight;
            let i_vector = point_light.world_transform[3].xyz - geometry.position;
            light.direction = normalize(i_vector);
            let light_distance = length(i_vector);
            light.color = srgb2physical(point_light.color.rgb) * point_light.intensity;
            light.color *= getDistanceAttenuation( light_distance, point_light.distance, point_light.decay );
            light.visible = any(light.color != vec3<f32>(0.0));
            return light;
        }
        $$ endif
        $$ if num_spot_lights > 0
        fn getSpotLightInfo( spot_light: SpotLight, geometry: GeometricContext ) -> IncidentLight {
            var light: IncidentLight;
            let i_vector = spot_light.world_transform[3].xyz - geometry.position;
            light.direction = normalize(i_vector);
            let angle_cos = dot(light.direction, -spot_light.direction.xyz);
            let spot_attenuation = getSpotAttenuation(spot_light.cone_cos, spot_light.penumbra_cos, angle_cos);
            if ( spot_attenuation > 0.0 ) {
                let light_distance = length( i_vector );
                light.color = srgb2physical(spot_light.color.rgb) * spot_light.intensity;
                light.color *= spot_attenuation;
                light.color *= getDistanceAttenuation( light_distance, spot_light.distance, spot_light.decay );
                light.visible = any(light.color != vec3<f32>(0.0));
            } else {
                light.color = vec3<f32>( 0.0 );
                light.visible = false;
            }
            return light;
        }
        $$ endif
        """

    def light_deps_pbr(self):
        return """
        struct PhysicalMaterial {
            diffuse_color: vec3<f32>,
            roughness: f32,
            specular_color: vec3<f32>,
            specular_f90: f32,
        };
        struct LightScatter {
            single_scatter: vec3<f32>,
            multi_scatter: vec3<f32>,
        };
        fn perturbNormal2Arb( eye_pos: vec3<f32>, surf_norm: vec3<f32>, mapN: vec3<f32>, uv: vec2<f32>, is_front: bool) -> vec3<f32> {
            let q0 = dpdx( eye_pos.xyz );
            let q1 = dpdy( eye_pos.xyz );
            let st0 = dpdx( uv.xy );
            let st1 = dpdy( uv.xy );
            let N = surf_norm; //  normalized
            let q1perp = cross( q1, N );
            let q0perp = cross( N, q0 );
            let T = q1perp * st0.x + q0perp * st1.x;
            let B = q1perp * st0.y + q0perp * st1.y;
            let det = max( dot( T, T ), dot( B, B ) );
            let faceDirection = f32(is_front) * 2.0 - 1.0;
            let scale = faceDirection * inverseSqrt(det);
            return normalize(T * mapN.x * scale + B * mapN.y * scale + N * mapN.z);
        }
        fn getMipLevel(maxMIPLevelScalar: f32, level: f32) -> f32 {
            let sigma = (3.141592653589793 * level * level) / (1.0 + level);
            let desiredMIPLevel = maxMIPLevelScalar + log2(sigma);
            let mip_level = clamp(desiredMIPLevel, 0.0, maxMIPLevelScalar);
            return mip_level;
        }
        fn getIBLIrradiance( normal: vec3<f32>, env_map: texture_cube<f32>, env_map_sampler: sampler, mip_level: f32) -> vec3<f32> {
            let envMapColor_srgb = textureSampleLevel( env_map, env_map_sampler, vec3<f32>( -normal.x, normal.yz), mip_level );
            return srgb2physical(envMapColor_srgb.rgb) * u_material.env_map_intensity * PI;
        }
        fn getIBLRadiance( view_dir: vec3<f32>, normal: vec3<f32>, roughness: f32, env_map: texture_cube<f32>, env_map_sampler: sampler, mip_level: f32 ) -> vec3<f32> {
            var reflectVec = reflect( - view_dir, normal );
            reflectVec = normalize(mix(reflectVec, normal, roughness*roughness));
            let envMapColor_srgb = textureSampleLevel( env_map, env_map_sampler, vec3<f32>( -reflectVec.x, reflectVec.yz), mip_level );
            return srgb2physical(envMapColor_srgb.rgb) * u_material.env_map_intensity;
        }
        fn computeMultiscattering(normal: vec3<f32>, view_dir: vec3<f32>, specular_color: vec3<f32>, specular_f90: f32, roughness: f32) -> LightScatter {
            let fab = DFGApprox( normal, view_dir, roughness );
            let FssEss = specular_color * fab.x + specular_f90 * fab.y;
            let Ess: f32 = fab.x + fab.y;
            let Ems: f32 = 1.0 - Ess;
            let Favg = specular_color + ( 1.0 - specular_color ) * 0.047619; // 1/21
            let Fms = FssEss * Favg / ( 1.0 - Ems * Favg );
            var scatter: LightScatter;
            scatter.single_scatter = FssEss;
            scatter.multi_scatter = Fms * Ems;
            return scatter;
        }
        fn RE_IndirectSpecular_Physical(radiance: vec3<f32>, irradiance: vec3<f32>,
                geometry: GeometricContext, material: PhysicalMaterial, reflected_light: ReflectedLight) -> ReflectedLight{
            let cosineWeightedIrradiance: vec3<f32> = irradiance * RECIPROCAL_PI;
            let scatter = computeMultiscattering( geometry.normal, geometry.view_dir, material.specular_color, material.specular_f90, material.roughness);
            //let total_scattering = scatter.single_scatter + scatter.multi_scatter;
            //let diffuse = material.diffuse_color * ( 1.0 - max( max( total_scattering.r, total_scattering.g ), total_scattering.b ) );
            let diffuse = material.diffuse_color * ( 1.0 - scatter.single_scatter - scatter.multi_scatter);
            var out_reflected_light: ReflectedLight = reflected_light;
            out_reflected_light.indirect_specular += (radiance * scatter.single_scatter + scatter.multi_scatter * cosineWeightedIrradiance);
            out_reflected_light.indirect_diffuse += diffuse * cosineWeightedIrradiance;
            return out_reflected_light;
        }
        fn RE_IndirectDiffuse_Physical(irradiance: vec3<f32>, geometry: GeometricContext, material: PhysicalMaterial, reflected_light: ReflectedLight) -> ReflectedLight {
            var out_reflected_light: ReflectedLight = reflected_light;
            out_reflected_light.indirect_diffuse += irradiance * BRDF_Lambert( material.diffuse_color );
            return out_reflected_light;
        }
        fn RE_Direct_Physical(direct_light: IncidentLight, geometry: GeometricContext, material: PhysicalMaterial, reflected_light: ReflectedLight) -> ReflectedLight {
            let dot_nl = clamp( dot( geometry.normal, direct_light.direction ), 0.0, 1.0 );
            let irradiance = dot_nl * direct_light.color;
            var out_reflected_light: ReflectedLight = reflected_light;
            out_reflected_light.direct_specular += irradiance * BRDF_GGX( direct_light.direction, geometry.view_dir, geometry.normal, material.specular_color, material.specular_f90, material.roughness );
            out_reflected_light.indirect_diffuse += irradiance * BRDF_Lambert( material.diffuse_color );
            return out_reflected_light;
        }
        fn RE_AmbientOcclusion_Physical(ambientOcclusion: f32, geometry: GeometricContext, material: PhysicalMaterial, reflected_light: ReflectedLight) -> ReflectedLight {
            let dot_nv = clamp( dot( geometry.normal, geometry.view_dir ), 0.0, 1.0);
            let ao_nv = dot_nv + ambientOcclusion;
            let ao_exp = exp2( -16.0 * material.roughness - 1.0 );
            let ao = clamp( pow(ao_nv, ao_exp) - 1.0 + ambientOcclusion, 0.0, 1.0 );
            var out_reflected_light: ReflectedLight = reflected_light;
            out_reflected_light.indirect_diffuse *= ambientOcclusion;
            out_reflected_light.indirect_specular *= ao;
            return out_reflected_light;
        }
    """

    def bsdfs_basic(self):
        # Bidirectional scattering distribution function
        return """
        fn BRDF_Lambert(diffuse_color: vec3<f32>) -> vec3<f32> {
            return RECIPROCAL_PI * diffuse_color;
        }
        fn F_Schlick(f0: vec3<f32>, f90: f32, dot_vh: f32,) -> vec3<f32> {
            // Optimized variant (presented by Epic at SIGGRAPH '13)
            // https://cdn2.unrealengine.com/Resources/files/2013SiggraphPresentationsNotes-26915738.pdf
            let fresnel = exp2( ( - 5.55473 * dot_vh - 6.98316 ) * dot_vh );
            return f0 * ( 1.0 - fresnel ) + ( f90 * fresnel );
        }
    """

    def bsdfs_pbr(self):
        return """
        fn V_GGX_SmithCorrelated(alpha: f32, dot_nl: f32, dot_nv: f32) -> f32 {
            let a2 = pow(alpha, 2.0);
            let gv = dot_nl * sqrt(a2 + (1.0-a2) * pow(dot_nv, 2.0));
            let gl = dot_nv * sqrt(a2 + (1.0-a2) * pow(dot_nl, 2.0 ));
            let epsilon = 1.0e-6;
            return 0.5/ max( gv+gl, epsilon);
        }
        fn D_GGX(alpha: f32, dot_nh: f32) -> f32 {
            let a2 = pow( alpha, 2.0 );
            let denom = pow(dot_nh, 2.0) * (a2 - 1.0) + 1.0;
            return RECIPROCAL_PI * a2/pow(denom, 2.0);
        }
        fn BRDF_GGX(light_dir: vec3<f32>, view_dir: vec3<f32>, normal: vec3<f32>, f0: vec3<f32>, f90: f32, roughness: f32) -> vec3<f32> {
            let alpha = pow( roughness, 2.0 );
            let half_dir = normalize( light_dir + view_dir );
            let dot_nl = clamp( dot( normal, light_dir ), 0.0, 1.0 );
            let dot_nv = clamp( dot( normal, view_dir ), 0.0, 1.0 );
            let dot_nh = clamp( dot( normal, half_dir ), 0.0, 1.0 );
            let dot_vh = clamp( dot( view_dir, half_dir ), 0.0, 1.0 );
            let F = F_Schlick( f0, f90, dot_vh);
            let V = V_GGX_SmithCorrelated( alpha, dot_nl, dot_nv );
            let D = D_GGX( alpha, dot_nh );
            return F * ( V * D );
        }
        fn DFGApprox( normal: vec3<f32>, view_dir: vec3<f32>, roughness: f32 ) -> vec2<f32>{
            let dot_nv = clamp( dot( normal, view_dir ), 0.0, 1.0);
            let c0 = vec4<f32>(- 1.0, - 0.0275, - 0.572, 0.022);
            let c1 = vec4<f32>(1.0, 0.0425, 1.04, - 0.04);
            let r = roughness * c0 + c1;
            let a004 = min( r.x * r.x, exp2( - 9.28 * dot_nv ) ) * r.x + r.y;
            let fab: vec2<f32> = vec2<f32>( - 1.04, 1.04 ) * a004 + r.zw;
            return fab;
        }
    """

    def shadow(self):
        return """
        fn get_shadow(t_shadow: texture_depth_2d_array, u_shadow_sampler: sampler_comparison, layer_index: i32, shadow_coords: vec4<f32>, bias: f32) -> f32 {
            if (shadow_coords.w <= 0.0) {
                return 1.0;
            }
            // compensate for the Y-flip difference between the NDC and texture coordinates
            let proj_coords = shadow_coords.xyz / shadow_coords.w;
            let flip_correction = vec2<f32>(0.5, -0.5);
            let light_local = proj_coords.xy * flip_correction + vec2<f32>(0.5, 0.5);
            if (light_local.x < 0.0 || light_local.x > 1.0 || light_local.y < 0.0 || light_local.y > 1.0) {
                return 1.0;
            }
            var depth:f32 = proj_coords.z - bias;
            depth = clamp(depth, 0.0, 1.0);
            var shadow: f32 = 0.0;
            shadow += textureSampleCompareLevel(t_shadow, u_shadow_sampler, light_local, layer_index, depth, vec2<i32>(0, 0));
            shadow += textureSampleCompareLevel(t_shadow, u_shadow_sampler, light_local, layer_index, depth, vec2<i32>(1, 0));
            shadow += textureSampleCompareLevel(t_shadow, u_shadow_sampler, light_local, layer_index, depth, vec2<i32>(0, 1));
            shadow += textureSampleCompareLevel(t_shadow, u_shadow_sampler, light_local, layer_index, depth, vec2<i32>(1, 1));
            shadow += textureSampleCompareLevel(t_shadow, u_shadow_sampler, light_local, layer_index, depth, vec2<i32>(-1, 0));
            shadow += textureSampleCompareLevel(t_shadow, u_shadow_sampler, light_local, layer_index, depth, vec2<i32>(0, -1));
            shadow += textureSampleCompareLevel(t_shadow, u_shadow_sampler, light_local, layer_index, depth, vec2<i32>(-1, -1));
            shadow += textureSampleCompareLevel(t_shadow, u_shadow_sampler, light_local, layer_index, depth, vec2<i32>(-1, 1));
            shadow += textureSampleCompareLevel(t_shadow, u_shadow_sampler, light_local, layer_index, depth, vec2<i32>(1, -1));
            shadow += textureSampleCompareLevel(t_shadow, u_shadow_sampler, light_local, layer_index, depth, vec2<i32>(2, 0));
            shadow += textureSampleCompareLevel(t_shadow, u_shadow_sampler, light_local, layer_index, depth, vec2<i32>(0, 2));
            shadow += textureSampleCompareLevel(t_shadow, u_shadow_sampler, light_local, layer_index, depth, vec2<i32>(2, 2));
            shadow += textureSampleCompareLevel(t_shadow, u_shadow_sampler, light_local, layer_index, depth, vec2<i32>(-2, 0));
            shadow += textureSampleCompareLevel(t_shadow, u_shadow_sampler, light_local, layer_index, depth, vec2<i32>(0, -2));
            shadow += textureSampleCompareLevel(t_shadow, u_shadow_sampler, light_local, layer_index, depth, vec2<i32>(-2, -2));
            shadow += textureSampleCompareLevel(t_shadow, u_shadow_sampler, light_local, layer_index, depth, vec2<i32>(-2, 2));
            shadow += textureSampleCompareLevel(t_shadow, u_shadow_sampler, light_local, layer_index, depth, vec2<i32>(2, -2));
            shadow /= 17.0;
            return shadow;
        }
        fn uv_to_direction( face:i32,  uv:vec2<f32>)->vec3<f32> {
            var u = 2.0 * uv.x - 1.0;
            var v = -2.0 * uv.y + 1.0;
            switch face{
                case 0: { return vec3<f32>(1.0, v, -u); }
                case 1: { return vec3<f32>(-1.0, v, u); }
                case 2: { return vec3<f32>(u, 1.0, -v); }
                case 3: { return vec3<f32>(u, -1.0, v); }
                case 4: { return vec3<f32>(u, v, 1.0); }
                case 5: { return vec3<f32>(-u, v, -1.0); }
                default: { return vec3<f32>(1.0, 1.0, 1.0); }
            }
        }
        fn get_cube_shadow(t_shadow: texture_depth_cube_array, u_shadow_sampler: sampler_comparison, layer_index: i32,
            light_view_proj: array<mat4x4<f32>,6>, world_pos: vec3<f32>, light_direction: vec3<f32>,  bias: f32) -> f32 {
            //var direction = world_pos - light_pos;
            var direction = -light_direction;
            let scale = 1.0 / max(max(abs(direction.x), abs(direction.y)), abs(direction.z));
            direction = direction * scale;
            let epsilon = 1.0e-6;
            var faceIndex = 0;
            var view_proj: mat4x4<f32>;
            if (abs(direction.x - 1.0) < epsilon) {
                faceIndex = 0;
                view_proj = light_view_proj[0];
            } else if (abs(direction.x + 1.0) < epsilon) {
                faceIndex = 1;
                view_proj = light_view_proj[1];
            }  else if (abs(direction.y - 1.0) < epsilon) {
                faceIndex = 2;
                view_proj = light_view_proj[2];
            } else if (abs(direction.y + 1.0) < epsilon) {
                faceIndex = 3;
                view_proj = light_view_proj[3];
            } else if (abs(direction.z - 1.0) < epsilon) {
                faceIndex = 4;
                view_proj = light_view_proj[4];
            } else if (abs(direction.z + 1.0) < epsilon) {
                faceIndex = 5;
                view_proj = light_view_proj[5];
            }
            var shadow_coords = view_proj * vec4<f32>(world_pos, 1.0);
            if (shadow_coords.w <= 0.0) {
                return 1.0;
            }
            let proj_coords = shadow_coords.xyz / shadow_coords.w;
            let flip_correction = vec2<f32>(0.5, -0.5);
            let light_local = proj_coords.xy * flip_correction + vec2<f32>(0.5, 0.5);
            var depth:f32 = proj_coords.z - bias; // bias?
            depth = clamp(depth, 0.0, 1.0);
            var dir = uv_to_direction(faceIndex, light_local);
            var shadow = textureSampleCompareLevel(t_shadow, u_shadow_sampler, dir, layer_index, depth);
            return shadow;
        }
    """

    def lighting_phong_simple(self):
        # Simple phong shading with builtin lights. What we used before #324
        return """
        fn lighting_phong(
            is_front: bool,
            varyings: Varyings,
            normal: vec3<f32>,
            view_dir: vec3<f32>,
            albeido: vec3<f32>,
        ) -> vec3<f32> {
            let light_color = srgb2physical(vec3<f32>(1.0, 1.0, 1.0));

            // Light parameters
            let ambient_factor = 0.1;
            let diffuse_factor = 0.7;
            let specular_factor = 0.3;
            let shininess = u_material.shininess;

            // Base vectors
            let view = normalize(view_dir);
            let light = view;
            var normal = select(-normal, normal, is_front);  // See pygfx/issues/#105 for details

            // Ambient
            let ambient_color = light_color * ambient_factor;

            // Diffuse (blinn-phong reflection model)
            let lambert_term = clamp(dot(light, normal), 0.0, 1.0);
            let diffuse_color = diffuse_factor * light_color * lambert_term;

            // Specular
            let halfway = normalize(light + view);  // halfway vector
            var specular_term = pow(clamp(dot(halfway,  normal), 0.0, 1.0), shininess);
            specular_term = select(0.0, specular_term, shininess > 0.0);
            let specular_color = specular_factor * specular_term * light_color;

            // Emissive color is additive and unaffected by lights
            let emissive_color = srgb2physical(u_material.emissive_color.rgb);

            // Put together
            return albeido * (ambient_color + diffuse_color) + specular_color + emissive_color;
        }
        """

    # %%%%% The actual entrypoints used by the shaders %%%%%

    def lighting_phong(self):
        # Phong shading using the lights present in the scene
        return (
            self.bsdfs_basic()
            + self.light_deps_basic()
            + """

        fn G_BlinnPhong_Implicit() -> f32 {
            return 0.25;
        }
        fn D_BlinnPhong(shininess: f32, dot_nh: f32) -> f32 {
            return RECIPROCAL_PI * ( shininess * 0.5 + 1.0 ) * pow( dot_nh, shininess );
        }
        fn BRDF_BlinnPhong(
            light_dir: vec3<f32>,
            view_dir: vec3<f32>,
            normal: vec3<f32>,
            specular_color: vec3<f32>,
            shininess: f32,
        ) -> vec3<f32> {
            let half_dir = normalize(light_dir + view_dir);
            let dot_nh = clamp(dot(normal, half_dir), 0.0, 1.0);
            let dot_vh = clamp(dot(view_dir, half_dir), 0.0, 1.0);
            let F = F_Schlick(specular_color, 1.0, dot_vh);
            let G = G_BlinnPhong_Implicit();
            let D = D_BlinnPhong(shininess, dot_nh);
            return F * ( G * D );
        }

        struct BlinnPhongMaterial {
            diffuse_color: vec3<f32>,
            specular_color: vec3<f32>,
            specular_shininess: f32,
            specular_strength: f32,
        };
        fn RE_Direct_BlinnPhong(
            direct_light: IncidentLight,
            geometry: GeometricContext,
            material: BlinnPhongMaterial,
            reflected_light: ReflectedLight,
        ) -> ReflectedLight {
            let dot_nl = clamp(dot(geometry.normal, direct_light.direction), 0.0, 1.0);
            let irradiance = dot_nl * direct_light.color;
            let direct_diffuse = irradiance * BRDF_Lambert( material.diffuse_color );
            let direct_specular = irradiance * BRDF_BlinnPhong( direct_light.direction, geometry.view_dir, geometry.normal, material.specular_color, material.specular_shininess ) * material.specular_strength;
            var out_reflected_light: ReflectedLight;
            out_reflected_light.direct_diffuse = reflected_light.direct_diffuse + direct_diffuse;
            out_reflected_light.direct_specular = reflected_light.direct_specular + direct_specular;
            out_reflected_light.indirect_diffuse = reflected_light.indirect_diffuse;
            out_reflected_light.indirect_specular = reflected_light.indirect_specular;
            return out_reflected_light;
        }
        fn RE_IndirectDiffuse_BlinnPhong(
            irradiance: vec3<f32>,
            geometry: GeometricContext,
            material: BlinnPhongMaterial,
            reflected_light: ReflectedLight,
        ) -> ReflectedLight {
            let indirect_diffuse = irradiance * BRDF_Lambert( material.diffuse_color );
            var out_reflected_light: ReflectedLight;
            out_reflected_light.direct_diffuse = reflected_light.direct_diffuse;
            out_reflected_light.direct_specular = reflected_light.direct_specular;
            out_reflected_light.indirect_diffuse = reflected_light.indirect_diffuse + indirect_diffuse;
            out_reflected_light.indirect_specular = reflected_light.indirect_specular;
            return out_reflected_light;
        }

        fn lighting_phong(
            varyings: Varyings,
            normal: vec3<f32>,
            view_dir: vec3<f32>,
            albeido: vec3<f32>,
        ) -> vec3<f32> {

            // Colors incoming via uniforms
            let specular_color = srgb2physical(u_material.specular_color.rgb);
            let ambient_color = u_ambient_light.color.rgb; // the one exception that is already physical

            var material: BlinnPhongMaterial;
            material.diffuse_color = albeido;
            material.specular_color = specular_color;
            material.specular_shininess = u_material.shininess;
            material.specular_strength = 1.0;   //  We could provide a specular map
            var reflected_light: ReflectedLight = ReflectedLight(vec3<f32>(0.0), vec3<f32>(0.0), vec3<f32>(0.0), vec3<f32>(0.0));

            var geometry: GeometricContext;
            geometry.position = varyings.world_pos;
            geometry.normal = normal;
            geometry.view_dir = view_dir;
            var i = 0;
            $$ if num_point_lights > 0
                loop {
                    if (i >= {{ num_point_lights }}) { break; }
                    let point_light = u_point_lights[i];
                    var light = getPointLightInfo(point_light, geometry);
                    if (! light.visible) { continue; }
                    $$ if receive_shadow
                    if (point_light.cast_shadow != 0){
                        let shadow = get_cube_shadow(u_shadow_map_point_light, u_shadow_sampler, i, point_light.light_view_proj_matrix, geometry.position, light.direction, point_light.shadow_bias);
                        light.color *= shadow;
                    }
                    $$ endif
                    reflected_light = RE_Direct_BlinnPhong( light, geometry, material, reflected_light );
                    continuing {
                        i += 1;
                    }
                }
            $$ endif
            $$ if num_spot_lights > 0
                i = 0;
                loop {
                    if (i >= {{ num_spot_lights }}) { break; }
                    let spot_light = u_spot_lights[i];
                    var light = getSpotLightInfo(spot_light, geometry);
                    if (! light.visible) { continue; }
                    $$ if receive_shadow
                    if (spot_light.cast_shadow != 0){
                        let coords = spot_light.light_view_proj_matrix * vec4<f32>(geometry.position,1.0);
                        let bias = spot_light.shadow_bias;
                        let shadow = get_shadow(u_shadow_map_spot_light, u_shadow_sampler, i, coords, bias);
                        light.color *= shadow;
                    }
                    $$ endif
                    reflected_light = RE_Direct_BlinnPhong( light, geometry, material, reflected_light );
                    continuing {
                        i += 1;
                    }
                }
            $$ endif
            $$ if num_dir_lights > 0
                i = 0;
                loop {
                    if (i >= {{ num_dir_lights }}) { break; }
                    let dir_light = u_directional_lights[i];
                    var light = getDirectionalLightInfo(dir_light, geometry);
                    if (! light.visible) { continue; }
                    $$ if receive_shadow
                    if (dir_light.cast_shadow != 0) {
                        let coords = dir_light.light_view_proj_matrix * vec4<f32>(geometry.position,1.0);
                        let bias = dir_light.shadow_bias;
                        let shadow = get_shadow(u_shadow_map_dir_light, u_shadow_sampler, i, coords, bias);
                        light.color *= shadow;
                    }
                    $$ endif
                    reflected_light = RE_Direct_BlinnPhong( light, geometry, material, reflected_light );
                    continuing {
                        i += 1;
                    }
                }
            $$ endif
            let irradiance = getAmbientLightIrradiance( ambient_color );
            reflected_light = RE_IndirectDiffuse_BlinnPhong( irradiance, geometry, material, reflected_light );
            return reflected_light.direct_diffuse + reflected_light.direct_specular + reflected_light.indirect_diffuse + reflected_light.indirect_specular + u_material.emissive_color.rgb;
        }
        """
        )

    def lighting_pbr(self):
        return (
            self.bsdfs_basic()
            + self.bsdfs_pbr()
            + self.light_deps_basic()
            + self.light_deps_pbr()
            + """

        fn lighting_pbr(
            varyings: Varyings,
            normal: vec3<f32>,
            view_dir: vec3<f32>,
            albeido: vec3<f32>,
        ) -> vec3<f32> {

            // Metalness
            var metalness_factor: f32 = u_material.metalness;
            $$ if use_metalness_map is defined
                metalness_factor *= textureSample( t_metalness_map, s_metalness_map, varyings.texcoord ).b;
            $$ endif

            // Roughness
            var roughness_factor: f32 = u_material.roughness;
            $$ if use_roughness_map is defined
                roughness_factor *= textureSample( t_roughness_map, s_roughness_map, varyings.texcoord ).g;
            $$ endif
            roughness_factor = max( roughness_factor, 0.0525 );
            let dxy = max( abs( dpdx( varyings.geometry_normal ) ), abs( dpdy( varyings.geometry_normal ) ) );
            let geometry_roughness = max( max( dxy.x, dxy.y ), dxy.z );

            // Define material
            var material: PhysicalMaterial;
            material.diffuse_color = albeido * ( 1.0 - metalness_factor );
            material.specular_color = mix( vec3<f32>( 0.04 ), albeido.rgb, metalness_factor );
            material.roughness = min( roughness_factor + geometry_roughness, 1.0 );
            material.specular_f90 = 1.0;



            // Define geometry
            var geometry: GeometricContext;
            geometry.position = varyings.world_pos;
            geometry.normal = normal;
            geometry.view_dir = view_dir;

            // Init the reflected light. Defines diffuse and specular, both direct and indirect
            var reflected_light: ReflectedLight = ReflectedLight(vec3<f32>(0.0), vec3<f32>(0.0), vec3<f32>(0.0), vec3<f32>(0.0));

            // Direct light from light sources
            var i = 0;
            $$ if num_point_lights > 0
                i = 0;
                loop {
                    if (i >= {{ num_point_lights }}) { break; }
                    let point_light = u_point_lights[i];
                    var light = getPointLightInfo(point_light, geometry);
                    if (! light.visible) { continue; }
                    $$ if receive_shadow
                    if (point_light.cast_shadow != 0){
                        let shadow = get_cube_shadow(u_shadow_map_point_light, u_shadow_sampler, i, point_light.light_view_proj_matrix, geometry.position, light.direction, point_light.shadow_bias);
                        light.color *= shadow;
                    }
                    $$ endif
                    reflected_light = RE_Direct_Physical( light, geometry, material, reflected_light );
                    continuing {
                        i += 1;
                    }
                }
            $$ endif
            $$ if num_spot_lights > 0
                i = 0;
                loop {
                    if (i >= {{ num_spot_lights }}) { break; }
                    let spot_light = u_spot_lights[i];
                    var light = getSpotLightInfo(spot_light, geometry);
                    if (! light.visible) { continue; }
                    $$ if receive_shadow
                    if (spot_light.cast_shadow != 0){
                        let coords = spot_light.light_view_proj_matrix * vec4<f32>(geometry.position,1.0);
                        let shadow = get_shadow(u_shadow_map_spot_light, u_shadow_sampler, i, coords, spot_light.shadow_bias);
                        light.color *= shadow;
                    }
                    $$ endif
                    reflected_light = RE_Direct_Physical( light, geometry, material, reflected_light );
                    continuing {
                        i += 1;
                    }
                }
            $$ endif
            $$ if num_dir_lights > 0
                i = 0;
                loop {
                    if (i >= {{ num_dir_lights }}) { break; }
                    let dir_light = u_directional_lights[i];
                    var light = getDirectionalLightInfo(dir_light, geometry);
                    if (! light.visible) { continue; }
                    $$ if receive_shadow
                    if (dir_light.cast_shadow != 0) {
                        let coords = dir_light.light_view_proj_matrix * vec4<f32>(geometry.position,1.0);
                        let shadow = get_shadow(u_shadow_map_dir_light, u_shadow_sampler, i, coords, dir_light.shadow_bias);
                        light.color *= shadow;
                    }
                    $$ endif
                    reflected_light = RE_Direct_Physical( light, geometry, material, reflected_light );
                    continuing {
                        i += 1;
                    }
                }
            $$ endif

            // The rest is for indirect light

            let ambient_color = u_ambient_light.color.rgb;  // the one exception that is already physical
            var irradiance = getAmbientLightIrradiance( ambient_color );

            // Light map (pre-backed lighting)
            $$ if use_light_map is defined
            let light_map_color = srgb2physical( textureSample( t_light_map, s_light_map, varyings.texcoord )).rgb );
            irradiance += light_map_color * u_material.light_map_intensity;
            // Note that if we implement light map for MeshBasicMaterial, we must multiply the intensity with the reciprocal PI.
            $$ endif

            // Process irradiance
            reflected_light = RE_IndirectDiffuse_Physical( irradiance, geometry, material, reflected_light );

            // Environment map (srgb2physical and intensity is handled in the getter functions)
            $$ if use_env_map is defined
            let mip_level_r = getMipLevel(u_material.env_map_max_mip_level, material.roughness);
            let ibl_radiance = getIBLRadiance( geometry.view_dir, geometry.normal, material.roughness, t_env_map, s_env_map, mip_level_r );
            let mip_level_i = getMipLevel(u_material.env_map_max_mip_level, 1.0);
            let ibl_irradiance = getIBLIrradiance( geometry.normal, t_env_map, s_env_map, mip_level_i );
            reflected_light = RE_IndirectSpecular_Physical(ibl_radiance, ibl_irradiance, geometry, material, reflected_light);
            $$ endif

            // Ambient occlusion
            $$ if use_ao_map is defined
            let ao_map_intensity = u_material.ao_map_intensity;
            let ambientOcclusion = ( textureSample( t_ao_map, s_ao_map, varyings.texcoord ).r - 1.0 ) * ao_map_intensity + 1.0;
            reflected_light = RE_AmbientOcclusion_Physical(ambientOcclusion, geometry, material, reflected_light);
            $$ endif

            // Combine direct and indirect light
            var lit_color = reflected_light.direct_diffuse + reflected_light.direct_specular + reflected_light.indirect_diffuse + reflected_light.indirect_specular;

            // Add emissive color
            var emissive_color = srgb2physical(u_material.emissive_color.rgb);
            $$ if use_emissive_map is defined
            emissive_color *= srgb2physical(textureSample(t_emissive_map, s_emissive_map, varyings.texcoord).rgb);
            $$ endif
            lit_color += emissive_color * u_material.emissive_intensity;

            return lit_color;
        }
        """
        )


shaderlib = Shaderlib()



import wgpu  # only for flags/enums

from . import register_wgpu_render_function, WorldObjectShader, Binding, RenderMask
from ...objects import Text
from ...materials import TextMaterial
from ...utils.text._shaper import REF_GLYPH_SIZE


@register_wgpu_render_function(Text, TextMaterial)
class TextShader(WorldObjectShader):
    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)
        geometry = wobject.geometry
        material = wobject.material
        self["screen_space"] = geometry.screen_space
        self["aa"] = material.aa

    def get_bindings(self, wobject, shared):
        geometry = wobject.geometry
        material = wobject.material

        sbuffer = "buffer/read_only_storage"
        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
            Binding("s_indices", sbuffer, geometry.indices, "VERTEX"),
            Binding("s_positions", sbuffer, geometry.positions, "VERTEX"),
            Binding("s_sizes", sbuffer, geometry.sizes, "VERTEX"),
        ]

        view = shared.glyph_atlas_texture_view
        bindings.append(Binding("s_atlas", "sampler/filtering", view, "FRAGMENT"))
        bindings.append(Binding("t_atlas", "texture/auto", view, "FRAGMENT"))

        # Let the shader generate code for our bindings
        bindings = {i: b for i, b in enumerate(bindings)}
        self.define_bindings(0, bindings)

        bindings1 = {}
        bindings1[0] = Binding(
            "s_glyph_infos", sbuffer, shared.glyph_atlas_info_buffer, "VERTEX"
        )

        return {
            0: bindings,
            1: bindings1,
        }

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_list,
            "cull_mode": wgpu.CullMode.none,
        }

    def get_render_info(self, wobject, shared):
        material = wobject.material
        n = wobject.geometry.positions.nitems * 6
        render_mask = wobject.render_mask
        if not render_mask:
            if material.is_transparent:
                render_mask = RenderMask.transparent
            elif material.color_is_transparent:
                render_mask = RenderMask.transparent
            else:
                render_mask = RenderMask.all
        return {
            "indices": (n, 1),
            "render_mask": render_mask,
        }

    def get_code(self):
        sizes = f"""
        let REF_GLYPH_SIZE: i32 = {REF_GLYPH_SIZE};
        """

        return (
            self.code_definitions()
            + sizes
            + self.code_common()
            + self.code_vertex()
            + self.code_fragment()
        )

    def code_vertex(self):
        return """

        struct VertexInput {
            @builtin(vertex_index) vertex_index : u32,
        };

        struct GlyphInfo {
            origin: vec2<i32>, // origin in the atlas
            size: vec2<i32>,  // size of the glyph (in the atlas)
            offset: vec2<f32>,  // positional offset of the glyph
        };
        @group(1) @binding(0)
        var<storage,read> s_glyph_infos: array<GlyphInfo>;

        @vertex
        fn vs_main(in: VertexInput) -> Varyings {

            let screen_factor = u_stdinfo.logical_size.xy / 2.0;

            let raw_index = i32(in.vertex_index);
            let index = raw_index / 6;
            let sub_index = raw_index % 6;

            // Load glyph info
            let glyph_index_raw = u32(load_s_indices(index));
            let font_size = load_s_sizes(index);
            let glyph_pos = load_s_positions(index);

            // Extract actual glyph index and the encoded font props
            let glyph_index = i32(glyph_index_raw & 0x00FFFFFFu);
            let weight_0_15 = (glyph_index_raw & 0xF0000000u) >> 28u;  // highest 4 bits
            let is_slanted = bool(glyph_index_raw & 0x08000000u);
            //let reserved1 = bool(glyph_index_raw & 0x04000000u);
            //let reserved2 = bool(glyph_index_raw & 0x02000000u);
            //let reserved3 = bool(glyph_index_raw & 0x01000000u);

            // Load meta-data of the glyph in the atlas
            let glyph_info = s_glyph_infos[glyph_index];
            let bitmap_rect = vec4<i32>(glyph_info.origin, glyph_info.size );

            // Prep correction vectors
            // The first offsets the rectangle to put it on the baseline/origin.
            // The second puts it at the end of the atlas-glyph rectangle.
            let pos_offset1 = glyph_info.offset / f32(REF_GLYPH_SIZE);
            let pos_offset2 = vec2<f32>(bitmap_rect.zw) / f32(REF_GLYPH_SIZE);

            var corners = array<vec2<f32>, 6>(
                vec2<f32>(0.0, 0.0), vec2<f32>(0.0, 1.0), vec2<f32>(1.0, 0.0),
                vec2<f32>(0.0, 1.0), vec2<f32>(1.0, 0.0), vec2<f32>(1.0, 1.0),
            );
            let corner = corners[sub_index];

            // Apply slanting, a.k.a. automated obliques, a.k.a. fake italics
            let slant_factor = f32(is_slanted) * 0.23;  // emperically selected based on NotoSans-Italic
            let slant = vec2<f32>(0.5 - corner.y, 0.0) * slant_factor;

            let pos_corner_factor = corner * vec2<f32>(1.0, -1.0);
            let vertex_pos = glyph_pos + (pos_offset1 + pos_offset2 * pos_corner_factor + slant) * font_size;
            let texcoord_in_pixels = vec2<f32>(bitmap_rect.xy) + vec2<f32>(bitmap_rect.zw) * corner;

            $$ if screen_space

                // We take the object's pos (model pos is origin), move to NDC, and apply the
                // glyph-positioning in logical screen coords. The text position is affected
                // by the world_transform, but the local scale and rotation do not affect the position.
                // We apply these seperately in screen space, so the user can scale and rotate the text that way.

                let raw_pos = vec3<f32>(0.0, 0.0, 0.0);
                let world_pos = u_wobject.world_transform * vec4<f32>(raw_pos, 1.0);
                let ndc_pos = u_stdinfo.projection_transform * u_stdinfo.cam_transform * world_pos;
                let vertex_pos_rotated_and_scaled = u_wobject.rot_scale_transform * vec4<f32>(vertex_pos, 0.0, 1.0);
                let delta_ndc = vertex_pos_rotated_and_scaled.xy / screen_factor;

                // Pixel scale is easy
                let atlas_pixel_scale = font_size / f32(REF_GLYPH_SIZE);

            $$ else

                // We take the glyph positions as model pos, move to world and then NDC.

                let raw_pos = vec4<f32>(vertex_pos, 0.0, 1.0);
                let world_pos = u_wobject.world_transform * raw_pos;
                let ndc_pos = u_stdinfo.projection_transform * u_stdinfo.cam_transform * world_pos;
                let delta_ndc = vec2<f32>(0.0, 0.0);

                // For the pixel scale, we first project points in x and y direction, and calculate
                // their distance in screen space. The smallest distance is used for scale.
                // In other words. we measure how out-of-plane the text is to determine the amount of aa.

                // Part of the out-of-plane skew might be due to intentional anisotropic scaling (stretched text).
                // Therefore we measure the intentional part here, so we can compensate below.
                // Stretched text still becomes somewhat jaggy or blurry, but not as much as it normally would.
                let sx = length(vec3<f32>(u_wobject.world_transform [0][0], u_wobject.world_transform [0][1], u_wobject.world_transform [0][2]));
                let sy = length(vec3<f32>(u_wobject.world_transform [1][0], u_wobject.world_transform [1][1], u_wobject.world_transform [1][2]));
                let aspect_scale = sqrt(sy / sx);
                let scale_correct = vec2<f32>(aspect_scale, 1.0 / aspect_scale);

                let full_matrix = u_stdinfo.projection_transform * u_stdinfo.cam_transform * u_wobject.world_transform;

                let atlas_pixel_dx = vec2<f32>(font_size / f32(REF_GLYPH_SIZE), 0.0) * scale_correct.x;
                let raw_pos_dx = vec4<f32>(vertex_pos + atlas_pixel_dx, 0.0, 1.0);
                let ndc_pos_dx = full_matrix * raw_pos_dx;

                let atlas_pixel_dy = vec2<f32>(0.0, font_size / f32(REF_GLYPH_SIZE)) * scale_correct.y;
                let raw_pos_dy = vec4<f32>(vertex_pos + atlas_pixel_dy, 0.0, 1.0);
                let ndc_pos_dy = full_matrix * raw_pos_dy;

                let screen_pos = (ndc_pos.xy / ndc_pos.w) * screen_factor;
                let screen_pos_dx = (ndc_pos_dx.xy / ndc_pos_dx.w) * screen_factor;
                let screen_pos_dy = (ndc_pos_dy.xy / ndc_pos_dy.w) * screen_factor;
                let atlas_pixel_scale = min(distance(screen_pos, screen_pos_dx), distance(screen_pos, screen_pos_dy));

            $$ endif

            var varyings: Varyings;
            varyings.position = vec4<f32>(ndc_pos.xy + delta_ndc * ndc_pos.w, ndc_pos.zw);
            varyings.world_pos = vec3<f32>(world_pos.xyz / world_pos.w);
            varyings.atlas_pixel_scale = f32(atlas_pixel_scale);
            varyings.texcoord_in_pixels = vec2<f32>(texcoord_in_pixels);
            varyings.weight_offset = f32(f32(weight_0_15) * 50.0 - 250.0);  // encodes -250..500 in steps of 50

            // Picking
            varyings.pick_idx = u32(index);
            varyings.glyph_coord = vec2<f32>(corner);

            return varyings;
        }
        """

    def code_fragment(self):
        return """

        fn _sdf_smoothstep( low : f32, high : f32, x : f32 ) -> f32 {
            let t = clamp( ( x - low ) / ( high - low ), 0.0, 1.0 );
            return t * t * ( 3.0 - 2.0 * t );
        }

        @fragment
        fn fs_main(varyings: Varyings) -> FragmentOutput {

            // Get the float texcoord
            let atlas_size = textureDimensions(t_atlas);
            let texcoord = varyings.texcoord_in_pixels  / vec2<f32>(atlas_size);

            // Sample the distance. A value of 0.5 represents the edge of the glyph,
            // with positive values representing the inside.
            let atlas_value = textureSample(t_atlas, s_atlas, texcoord).r;

            // Convert to a more useful measure, where the edge is at 0.0, and the inside is negative.
            // The maximum value at which we can still detect the edge is just below 0.5.
            let distance = (0.5 - atlas_value);

            // Load tickness factors
            let weight_offset = clamp(varyings.weight_offset + u_material.weight_offset, -400.0, 1600.0);
            let weight_thickness = weight_offset * 0.00031;  // emperically derived factor
            let outline_thickness = u_material.outline_thickness;

            // The softness is calculated from the scale of one atlas-pixel in screen space.
            let max_softness = 0.75;
            let softness = clamp(0.0, max_softness, 2.0 / (f32(REF_GLYPH_SIZE) * varyings.atlas_pixel_scale));

            // Turns out that how thick a font looks depends on a number of factors:
            // - In pygfx the size of the font for which the sdf was created affects the output a bit.
            // - In a browser, the type of browser seems to affect the output a bit.
            // - In a browser, the OS matters more (e.g. Windows and MacOS handle aa and pixel alignment differently).
            // - The blurry edge for aa affects the perception of the weight.
            // - White text on black looks more bold than black text on white!
            //
            // Below you see how I gave the cut_off an offset that scales with the softness.
            // This might suggest that it compensates mostly for the aa-effect but that might be a
            // coincidence. All I did was try to bring our result close to the output of the
            // same text, rendered in a browser, where the text is white on a dark bg (I
            // checked against Firefox on MacOS, with retina display). Note that modern
            // browsers compensate for the white-on-dark effect. We cannot, because we don't
            // know whats behind the text, but the user can use weight_offset when the text
            // is darker than the bg. More info at issue #358.
            let cut_off_correction = 0.25 * softness;

            // Calculate cut-off's. Apply min so it's always a valid shape.
            let cut_off = min(0.49, 0.0 + cut_off_correction + weight_thickness + outline_thickness);
            let outline_cutoff = max(0.0, cut_off - outline_thickness);

            // Init opacity value to get the shape of the glyph
            var aa_alpha = 1.0;
            var soften_alpha = 1.0;
            var outline = 0.0;

            $$ if aa
                // We use smoothstep to include alpha blending.
                let outside_ness = _sdf_smoothstep(cut_off - softness, cut_off + softness, distance);
                aa_alpha = (1.0 - outside_ness);
                // High softness values also result in lower alpha to prevent artifacts under high angles.
                soften_alpha = 1.0 - max(softness / max_softness - 0.1, 0.0);
                // Outline
                let outline_softness = min(softness, 0.5 * outline_thickness);
                outline = _sdf_smoothstep(outline_cutoff - outline_softness, outline_cutoff + outline_softness, distance);
            $$ else
                // Do a hard transition
                aa_alpha = select(0.0, 1.0, distance < cut_off);
                outline = select(1.0, 0.0, distance < outline_cutoff);
            $$ endif

            // Early exit
            if (aa_alpha <= 0.0) { discard; }

            // For aa we reduce alpha quicker, which looks better to the human eye
            aa_alpha = aa_alpha * aa_alpha;

            // Turn outline really off if not used. Otherwise some of it may leak through in the aa fragments.
            outline = select(0.0, outline, outline_thickness > 0.0);

            // Get color for this fragment
            let base_srgb = u_material.color;
            let outline_srgb = u_material.outline_color;
            let color = mix(srgb2physical(base_srgb.rgb), srgb2physical(outline_srgb.rgb), outline);
            let color_alpha = mix(base_srgb.a, outline_srgb.a, outline);

            // Compose total opacity and the output color
            let opacity = u_material.opacity * color_alpha * aa_alpha * soften_alpha;
            var color_out = vec4<f32>(color, opacity);

            // Debug
            //color_out = vec4<f32>(atlas_value, 0.0, 0.0, 1.0);

            // Wrap up
            apply_clipping_planes(varyings.world_pos);
            var out = get_fragment_output(varyings.position.z, color_out);

            $$ if write_pick
            // The wobject-id must be 20 bits. In total it must not exceed 64 bits.
            out.pick = (
                pick_pack(u32(u_wobject.id), 20) +
                pick_pack(varyings.pick_idx, 26) +
                pick_pack(u32(varyings.glyph_coord.x * 512.0), 9) +
                pick_pack(u32(varyings.glyph_coord.y * 512.0), 9)
            );
            $$ endif

            return out;
        }
        """



from ._base import Material
from ..resources import TextureView
from ..utils import unpack_bitfield, Color


class LineMaterial(Material):
    """Basic line material.

    Parameters
    ----------
    color : Color
        The uniform color of the line. Ignored if ``vertex_colors`` is True.
    thickness : float
        The line thickness expressed in logical pixels.
    vertex_colors : bool
        Whether to use the vertex colors provided in the geometry to color the
        line. If True, color will be ignored.
    map : TextureView
        The texture map specifying the color for each texture coordinate. The
        dimensionality of the map can be 1D, 2D or 3D, but should match the
        number of columns in the geometry's texcoords.
    aa : bool
        Whether or not the line should be anti-aliased. Aliasing gives prettier
        results, but may affect performance for very large datasets. Default
        True.
    kwargs : Any
        Additional kwargs will be passed to the :class:`material base class
        <pygfx.Material>`.

    """

    uniform_type = dict(
        color="4xf4",
        thickness="f4",
    )

    def __init__(
        self,
        color=(1, 1, 1, 1),
        thickness=2.0,
        vertex_colors=False,
        map=None,
        aa=True,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.color = color
        self.aa = aa
        self.map = map
        self.thickness = thickness
        self._vertex_colors = bool(vertex_colors)

    def _wgpu_get_pick_info(self, pick_value):
        # This should match with the shader
        values = unpack_bitfield(pick_value, wobject_id=20, index=26, coord=18)
        return {
            "vertex_index": values["index"],
            "segment_coord": (values["coord"] - 100000) / 100000.0,
        }

    @property
    def color(self):
        """The uniform color of the line."""
        return Color(self.uniform_buffer.data["color"])

    @color.setter
    def color(self, color):
        color = Color(color)
        self.uniform_buffer.data["color"] = color
        self.uniform_buffer.update_range(0, 1)
        self._store.color_is_transparent = color.a < 1

    @property
    def color_is_transparent(self):
        """Whether the color is (semi) transparent (i.e. not fully opaque)."""
        return self._store.color_is_transparent

    @property
    def aa(self):
        """Whether or not the line should be anti-aliased. Aliasing
        gives prettier results, but may affect performance for very large
        datasets. Default True.
        """
        return self._store.aa

    @aa.setter
    def aa(self, aa):
        self._store.aa = bool(aa)

    @property
    def vertex_colors(self):
        """Whether to use the vertex colors provided in the geometry."""
        return self._vertex_colors

    @vertex_colors.setter
    def vertex_colors(self, value):
        value = bool(value)
        if value != self._vertex_colors:
            self._vertex_colors = value

    @property
    def thickness(self):
        """The line thickness expressed in logical pixels."""
        return float(self.uniform_buffer.data["thickness"])

    @thickness.setter
    def thickness(self, thickness):
        self.uniform_buffer.data["thickness"] = thickness
        self.uniform_buffer.update_range(0, 1)

    @property
    def map(self):
        """The texture map specifying the color for each texture coordinate.
        The dimensionality of the map can be 1D, 2D or 3D, but should match the
        number of columns in the geometry's texcoords.
        """
        return self._map

    @map.setter
    def map(self, map):
        assert map is None or isinstance(map, TextureView)
        self._map = map


class LineThinMaterial(LineMaterial):
    """Thin line material.

    A simple line, drawn with line_strip primitives that has a thickness
    of one physical pixel (the thickness property is ignored).

    While you probably don't want to use this property in your application (its
    width is inconsistent and looks *very* thin on HiDPI monitors), it can be
    useful for debugging as it is more performant than other line materials.

    """


class LineThinSegmentMaterial(LineMaterial):
    """Thin line segment material.

    Simple line segments, drawn with line primitives that has a thickness
    of one physical pixel (the thickness property is ignored).

    While you probably don't want to use this property in your application (its
    width is inconsistent and looks *very* thin on HiDPI monitors), it can be
    useful for debugging as it is more performant than other line materials.

    """


class LineSegmentMaterial(LineMaterial):
    """Line segment material.

    A material that renders line segments between each two subsequent points."""


class LineArrowMaterial(LineSegmentMaterial):
    """Arrow (vector) line material.

    A material that renders line segments that look like little vectors."""



from ..resources import TextureView
from ._base import Material


class ImageBasicMaterial(Material):
    """Rasterized image material.

    Parameters
    ----------
    clim : tuple
        The contrast limits to scale the data values with. Default (0, 1).
    map : TextureView
        The colormap to turn the image values into its final color. If not given
        or None, the values themselves represents the color. The dimensionality
        of the map can be 1D, 2D or 3D, but should match the number of channels
        in the image.
    kwargs : Any
        Additional kwargs will be passed to the :class:`material base class
        <pygfx.Material>`.

    """

    uniform_type = dict(
        clim="2xf4",
    )

    def __init__(self, clim=None, map=None, **kwargs):
        super().__init__(**kwargs)
        self.map = map
        self.clim = clim

    @property
    def map(self):
        """The colormap to turn the image values into its final color.
        If not given or None, the values themselves represents the color.
        The dimensionality of the map can be 1D, 2D or 3D, but should match the
        number of channels in the image.
        """
        return self._map

    @map.setter
    def map(self, map):
        assert map is None or isinstance(map, TextureView)
        self._map = map

    @property
    def clim(self):
        """The contrast limits to scale the data values with. Default (0, 1)."""
        v1, v2 = self.uniform_buffer.data["clim"]
        return float(v1), float(v2)

    @clim.setter
    def clim(self, clim):
        # Check and store given clim
        if clim is None:
            clim = 0, 1
        clim = float(clim[0]), float(clim[1])
        # Update uniform data
        self.uniform_buffer.data["clim"] = clim
        self.uniform_buffer.update_range(0, 1)



"""Helpers for visual debugging of scenes.

This module contains a collection of WorldObjects that can be useful
when debugging a scene or to create reference points within a scene.


.. currentmodule:: pygfx.helpers

.. autosummary::
    :toctree: helpers/
    :template: ../_templates/custom_layout.rst

    AxesHelper
    GridHelper
    BoxHelper
    TransformGizmo
    PointLightHelper
    DirectionalLightHelper
    SpotLightHelper

"""

# flake8: noqa

from ._axes import AxesHelper
from ._grid import GridHelper
from ._box import BoxHelper
from ._gizmo import TransformGizmo
from ._lights import (
    PointLightHelper,
    DirectionalLightHelper,
    SpotLightHelper,
)
